@@ -500,12 +500,18 @@ impl<R: Read + io::Seek> ZipArchive<R> {
     }
 
     /// Search for a file entry by name, decrypt with given password
-    /// 
+    ///
     /// # Warning
-    /// 
+    ///
     /// The implementation of the cryptographic algorithms has not
     /// gone through a correctness review, and you should assume it is insecure:
     /// passwords used with this API may be compromised.
+    ///
+    /// This function sometimes accepts wrong password. This is because the ZIP spec only allows us
+    /// to check for a 1/256 chance that the password is correct.
+    /// There are many passwords out there that will also pass the validity checks
+    /// we are able to perform. This is a weakness of the ZipCrypto algorithm,
+    /// due to its fairly primitive approach to cryptography.
     pub fn by_name_decrypt<'a>(
         &'a mut self,
         name: &str,
@@ -534,12 +540,18 @@ impl<R: Read + io::Seek> ZipArchive<R> {
     }
 
     /// Get a contained file by index, decrypt with given password
-    /// 
+    ///
     /// # Warning
-    /// 
+    ///
     /// The implementation of the cryptographic algorithms has not
     /// gone through a correctness review, and you should assume it is insecure:
     /// passwords used with this API may be compromised.
+    ///
+    /// This function sometimes accepts wrong password. This is because the ZIP spec only allows us
+    /// to check for a 1/256 chance that the password is correct.
+    /// There are many passwords out there that will also pass the validity checks
+    /// we are able to perform. This is a weakness of the ZipCrypto algorithm,
+    /// due to its fairly primitive approach to cryptography.
     pub fn by_index_decrypt<'a>(
         &'a mut self,
         file_number: usize,
@@ -679,6 +691,7 @@ pub(crate) fn central_header_to_zip_file<R: Read + io::Seek>(
             #[allow(deprecated)]
             CompressionMethod::from_u16(compression_method)
         },
+        compression_level: None,
         last_modified_time: DateTime::from_msdos(last_mod_date, last_mod_time),
         crc32,
         compressed_size: compressed_size as u64,
@@ -726,17 +739,17 @@ fn parse_extra_field(file: &mut ZipFileData) -> ZipResult<()> {
         match kind {
             // Zip64 extended information extra field
             0x0001 => {
-                if file.uncompressed_size == 0xFFFFFFFF {
+                if file.uncompressed_size == spec::ZIP64_BYTES_THR {
                     file.large_file = true;
                     file.uncompressed_size = reader.read_u64::<LittleEndian>()?;
                     len_left -= 8;
                 }
-                if file.compressed_size == 0xFFFFFFFF {
+                if file.compressed_size == spec::ZIP64_BYTES_THR {
                     file.large_file = true;
                     file.compressed_size = reader.read_u64::<LittleEndian>()?;
                     len_left -= 8;
                 }
-                if file.header_start == 0xFFFFFFFF {
+                if file.header_start == spec::ZIP64_BYTES_THR {
                     file.header_start = reader.read_u64::<LittleEndian>()?;
                     len_left -= 8;
                 }
@@ -1074,6 +1087,7 @@ pub fn read_zipfile_from_stream<'a, R: io::Read>(
         encrypted,
         using_data_descriptor,
         compression_method,
+        compression_level: None,
         last_modified_time: DateTime::from_msdos(last_mod_date, last_mod_time),
         crc32,
         compressed_size: compressed_size as u64,
